--- chan_zap.c.ORIG	2004-06-18 19:15:01.000000000 +0200
+++ chan_zap.c	2004-06-19 14:46:16.000000000 +0200
@@ -532,6 +532,7 @@
 	int alreadyhungup;
 	int proceeding;
 	int setup_ack;		/* wheter we received SETUP_ACKNOWLEDGE or not */
+	char pendingdigits[AST_MAX_EXTENSION];  /* Data held pending SETUP_ACKNOWLEDGE */
 #endif	
 #ifdef ZAPATA_R2
 	int r2prot;
@@ -846,9 +847,24 @@
 	index = zt_get_index(ast, p, 0);
 	if (index == SUB_REAL) {
 #ifdef ZAPATA_PRI
-		if (p->sig == SIG_PRI && ast->_state == AST_STATE_DIALING && p->setup_ack && (p->proceeding < 2)) {
+		if (p->sig == SIG_PRI && ast->_state == AST_STATE_DIALING && (p->proceeding < 2)) {
+			// Take the pri lock and the channel lock in the same order as
+			// in pri_dchannel (PRI_EVENT_SETUP_ACK) to prevent races.
 			if (!pri_grab(p, p->pri)) {
-				pri_information(p->pri->pri,p->call,digit);
+				ast_mutex_lock(&p->lock);
+				if (p->setup_ack) {
+					pri_information(p->pri->pri,p->call,digit);
+				} else {
+					int numpending;
+					numpending = strlen(p->pendingdigits);
+					if (numpending < (AST_MAX_EXTENSION-1) ) {
+						p->pendingdigits[numpending] = digit;
+						p->pendingdigits[numpending+1] = 0;
+					} else {
+						ast_log(LOG_WARNING, "Too many digits pending SETUP_ACK on span %d\n", p->span);
+					}
+				}
+				ast_mutex_unlock(&p->lock);
 				pri_rel(p->pri);
 			} else
 				ast_log(LOG_WARNING, "Unable to grab PRI on span %d\n", p->span);
@@ -2019,6 +2035,7 @@
 #ifdef ZAPATA_PRI
 		p->proceeding = 0;
 		p->setup_ack = 0;
+		p->pendingdigits[0] = 0;
 #endif		
 		if (p->dsp) {
 			ast_dsp_free(p->dsp);
@@ -7469,6 +7486,14 @@
 				} else {
 					ast_mutex_lock(&pri->pvts[chanpos]->lock);
 					pri->pvts[chanpos]->setup_ack = 1;
+					/* Send pending digits */
+					char * digitpos = pri->pvts[chanpos]->pendingdigits;
+					while (*digitpos) {
+						ast_log(LOG_DEBUG, "Sending pending digit %c\n", *digitpos);
+						pri_information(pri->pri,pri->pvts[chanpos]->call,*digitpos);
+						++digitpos;
+					}
+					pri->pvts[chanpos]->pendingdigits[0] = 0;
 					ast_mutex_unlock(&pri->pvts[chanpos]->lock);
 				}
 				break;
